@@ -4,46 +4,36 @@ import hashlib
 import re
 import requests
 import urllib
+from urllib.parse import urlparse
+from urllib.parse import urlunparse
 
 import dns.resolver
 import dns.dnssec
+from dns.resolver import Resolver
 from cryptography import x509
 from cryptography.hazmat.backends import default_backend
-from cryptography.hazmat.primitives import serialization
-from cryptography.hazmat.primitives.asymmetric import padding
-from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
-from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
-from cryptography.exceptions import InvalidSignature
-from cryptography.exceptions import UnsupportedAlgorithm
+from forcediphttpsadapter.adapters import ForcedIPHTTPSAdapter
 
 from .exceptions import TLSAError
+from .pki import PKI
 
 
 class DANE:
     """Abstract interactions involved in certificate retrieval."""
 
-    @classmethod
-    def build_x509_object(cls, certificate):
-        """Wrap format determination and return an x509 object.
-
-        Args:
-            certificate (str): Certificate in PEM or DER format.
-
-        Return:
-            cryptography.x509.Certificate object.
-
-        Raise:
-            TLSAError if unable to parse.
-        """
-        if isinstance(certificate, str):
-            certificate = certificate.encode()
-        try:
-            return x509.load_pem_x509_certificate(certificate,
-                                                  default_backend())
-        except ValueError:  # This hits if it's a DER cert.
-            pass
-        return x509.load_der_x509_certificate(certificate,
-                                              default_backend())
+    descr = {"certificate_usage":
+             {0: "CA Constraint",
+              1: "PKIX-EE",
+              2: "DANE-TA",
+              3: "DANE-EE",
+              4: "PKIX-CD"},
+             "selector":
+             {0: "Full certificate match",
+              1: "Public key match"},
+             "matching_type":
+             {0: "Exact match against certificate association",
+              1: "Match the SHA-256 hash of the selected content",
+              2: "Match the SHA-512 hash of the selected content"}}
 
     @classmethod
     def generate_tlsa_record(cls, certificate_usage, selector,
@@ -66,8 +56,8 @@ class DANE:
         Raise:
             TLSAError if unsupported options are used.
         """
-        x509_obj = cls.build_x509_object(certificate)
-        cert_bytes = x509_obj.public_bytes(encoding=serialization.Encoding.DER)
+        x509_obj = PKI.build_x509_object(certificate)
+        cert_bytes = PKI.serialize_cert(x509_obj, "DER")
         if matching_type == 0:
             certificate_association = binascii.hexlify(cert_bytes).decode()
         elif matching_type == 1:
@@ -103,22 +93,20 @@ class DANE:
         valid_selectors = [0, 1]
         valid_hashing_algos = {"sha256": hashlib.sha256,
                                "sha512": hashlib.sha512}
-        x509_obj = cls.build_x509_object(certificate)
+        x509_obj = PKI.build_x509_object(certificate)
         if selector not in valid_selectors:
             raise ValueError("Invalid selector.")
         if sha not in valid_hashing_algos:
             raise ValueError("Invalid sha.")
         if selector == 0:
-            hashable = x509_obj.public_bytes(serialization.Encoding.DER)
+            hashable = PKI.serialize_cert(x509_obj, "DER")
         elif selector == 1:
-            pubkey = x509_obj.public_key()
-            hashable = pubkey.public_bytes(serialization.Encoding.DER,
-                                           serialization.PublicFormat.SubjectPublicKeyInfo)  # NOQA
+            hashable = PKI.serialize_cert(x509_obj, "RPK_DER")
         hash = valid_hashing_algos[sha](hashable).hexdigest()
         return hash
 
     @classmethod
-    def get_first_leaf_certificate(cls, dnsname):
+    def get_first_leaf_certificate(cls, dnsname, nsaddr=None, dns_timeout=5):
         """Return the first leaf certificate from TLSA records at ``dnsname``.
 
         This method essentially wraps
@@ -128,20 +116,22 @@ class DANE:
 
         Args:
             dnsname (str): DNS name to query for certificate.
+            nsaddr (str): Override system resolver.
+            dns_timeout (int): Timeout in seconds for DNS query.
 
         Return:
             dict: Dictionary with keys for ``certificate_usage``, ``selector``,
                 ``matching_type``, ``certificate_association``.
                 If no leaf certificate is found, None is returned.
         """
-        all_tlsa_records = cls.get_tlsa_records(dnsname)
+        all_tlsa_records = cls.get_tlsa_records(dnsname, nsaddr=nsaddr, dns_timeout=dns_timeout)
         for tlsa in all_tlsa_records:
             if (tlsa["certificate_usage"] in [1, 3, 4]
                     and tlsa["matching_type"] == 0):
                 return tlsa
 
     @classmethod
-    def get_tlsa_records(cls, dnsname, nsaddr=None):
+    def get_tlsa_records(cls, dnsname, nsaddr=None, dns_timeout=5):
         """TLSA records in a list of dictionaries.
 
         This method retrieves and parses the TLSA records from
@@ -150,6 +140,7 @@ class DANE:
         Args:
             dnsname (str): DNS name to query for TLSA record.
             nsaddr (str): Nameserver address.
+            dns_timeout (int): Timeout in seconds for DNS query.
 
         Return:
             list of dict: Dictionaries with the following keys:
@@ -172,36 +163,37 @@ class DANE:
             # If the payload is actually a certificate, confirm that it parses.
             if result["matching_type"] == 0:
                 cert = result["certificate_association"]
-                cls.validate_certificate(cert)
+                PKI.validate_certificate_association(cert)
             for ctx in ["dnssec", "tcp", "tls"]:
                 result[ctx] = query_details[ctx]
             results.append(result.copy())
         return results
 
     @classmethod
-    def get_responses(cls, dnsname, rr_type, nsaddr=None):
+    def get_responses(cls, dnsname, rr_type, nsaddr=None, dns_timeout=5):
         """Return a list of dicts containing DNS RRs and security context.
 
         Args:
             dnsname (str): DNS name for query.
             rr_type (str): RR type to query. Defaults to TLSA.
+            nsaddr (str): Nameserver override address.
+            dns_timeout (int): Timeout in seconds for DNS query.
 
         Return:
             dict: Keys are ``responses`` (list of string),
                 ``dnssec`` (bool), ``tls`` (bool), ``tcp`` (bool).
         """
-        timeout = 5
-        default_recursor = dns.resolver.get_default_resolver().nameservers[0]
-        resolver = nsaddr if nsaddr else default_recursor
+        default_resolver = dns.resolver.get_default_resolver().nameservers[0]
+        resolver = nsaddr if nsaddr else default_resolver
         query = dns.message.make_query(dnsname, rr_type, want_dnssec=True)
         query_details = {"tls": False, "dnssec": False}
         try:
-            resp_msg = dns.query.tls(query, resolver, timeout=timeout)
+            resp_msg = dns.query.tls(query, resolver, timeout=dns_timeout)
             query_details["tls"] = True
             query_details["tcp"] = True
         except ConnectionRefusedError:
             resp_msg, was_tcp = dns.query.udp_with_fallback(query, resolver,
-                                                            timeout=timeout)
+                                                            timeout=dns_timeout)
             query_details["tcp"] = True if was_tcp else False
         flags_int = resp_msg.flags
         flags = dns.flags.to_text(flags_int)
@@ -213,30 +205,20 @@ class DANE:
         return query_details
 
     @classmethod
-    def certificate_association_to_der(cls, certificate_association):
-        """Return DER bytes from a TLSA record's ``certificate_association``.
-
-        Args:
-            certificate_association (str): Certificate association information
-                extracted from a TLSA record.
-
-        Return:
-            bytes: DER-formatted certificate.
-        """
-        return binascii.unhexlify(certificate_association)
-
-    @classmethod
-    def der_to_pem(cls, der_cert):
-        """Return the PEM representation of a TLSA certificate_association.
-
-        Args:
-            der (str): A certificate in DER format.
-
-        Return:
-            bytes: PEM-formatted certificate.
-        """
-        cert = cls.build_x509_object(der_cert)
-        return cert.public_bytes(serialization.Encoding.PEM)
+    def get_a_record(cls, dnsname, nsaddr=None, dns_timeout=5):
+        """Get the first A record."""
+        resolver = Resolver()
+        if nsaddr:
+            resolver.nameservers = [nsaddr]
+        try:
+            canonical_name = resolver.canonical_name(dnsname)
+            ip_address = str(resolver.resolve(canonical_name, "A")[0])
+        except dns.exception.DNSException as err:
+            msg = "Caught error '{}' when retrieving A record.".format(err)
+            raise ValueError(msg)
+        except IndexError:
+            raise ValueError("No A records for {}".format(dnsname))
+        return ip_address
 
     @classmethod
     def process_response(cls, response):
@@ -288,31 +270,16 @@ class DANE:
             issues.append("invalid matching type value")
         if not re.match("^[A-Za-z0-9]+$", tlsa_fields["certificate_association"]):
             issues.append("invalid certificate association value")
+        if int(tlsa_fields["matching_type"]) == 0:
+            try:
+                PKI.validate_certificate_association(tlsa_fields["certificate_association"])
+            except TLSAError as err:
+                issues.append(str(err))
         if issues:
             raise TLSAError("Malformed DNS record: {}.".format(", ".join(issues)))
 
     @classmethod
-    def validate_certificate(cls, certificate):
-        """Raise TLSAError if certificate does not parse, or return None.
-
-        Args:
-            certificate (str): Certificate association data from TLSA record.
-
-        Return:
-            None
-
-        Raise:
-            TLSAError if parsing fails.
-        """
-        try:
-            der = cls.certificate_association_to_der(certificate)
-            cls.build_x509_object(der)
-        except (ValueError, binascii.Error) as err:
-            msg = "Caught error '{}' with TLSA record.".format(err)
-            raise TLSAError(msg)
-
-    @classmethod
-    def authenticate_tlsa(cls, dns_name, record):
+    def authenticate_tlsa(cls, dns_name, record, nsaddr=None, dns_timeout=5):
         """Return None if the identity is authenticated, or raise ValueError.
 
         This method authenticates a TLSA record as follows:
@@ -341,6 +308,8 @@ class DANE:
             record (dict): Keys for ``certificate_usage``, ``selector``, 
                 ``matching_type``, ``certificate_association``, 
                 and ``dnssec``.
+            nsaddr (str): Name server override.
+            dns_timeout (int): Timeout in seconds for DNS query.
         
         Return:
             None if this identity can be cryptographically authenticated.
@@ -363,16 +332,17 @@ class DANE:
         if errmsg:
             raise TLSAError(errmsg)
         certificate_association = record["certificate_association"]
-        certificate_der = cls.certificate_association_to_der(certificate_association)
-        id_cert = cls.der_to_pem(certificate_der)
+        certificate_der = PKI.certificate_association_to_der(certificate_association)
+        id_cert = PKI.der_to_pem(certificate_der)
         try:
             # print("Load CA cert")
-            ca_cert = cls.get_ca_certificate_for_identity(dns_name, id_cert)
+            ca_certs = cls.get_ca_certificates_for_identity(dns_name, id_cert, nsaddr=nsaddr)
             # print("Test cert sig.")
-            if not cls.verify_certificate_signature(id_cert, ca_cert):
-                errmsg += "PKIX signature validation for identity failed.\n"
+            pkix_valid, reason = PKI.validate_certificate_chain(id_cert, ca_certs)
+            if not pkix_valid:
+                errmsg += "PKIX signature validation for identity failed: {}.\n".format(reason)
             # print("verify DNS name")
-            if not cls.verify_dnsname(dns_name, certificate_der):
+            if not PKI.verify_dnsname(dns_name, certificate_der):
                 errmsg += "DNS name match against SAN failed.\n"
         except TLSAError as err:
             errmsg += "Failed to get certificate for identity: {}\n".format(err)
@@ -380,61 +350,6 @@ class DANE:
             raise TLSAError(errmsg)
         return
 
-    @classmethod
-    def verify_dnsname(cls, dns_name, certificate_der):
-        """Return True if the first dNSName in the SAN matches."""
-        x5_obj = cls.build_x509_object(certificate_der)
-        san_dns_names = cls.get_dnsnames_from_cert(x5_obj)
-        if san_dns_names[0] != dns_name:
-            return False
-        return True
-
-    @classmethod
-    def get_dnsnames_from_cert(cls, x5_obj):
-        """Return the dnsnames from the certificate's SAN.
-
-        Args:
-            x5_obj (cryptography.x509): Certificate object.
-        
-        Return: 
-            list: str: dNSNames from certificate SAN.
-        """
-        san = x5_obj.extensions.get_extension_for_class(x509.SubjectAlternativeName)
-        return san.value.get_values_for_type(x509.DNSName)
-
-    @classmethod
-    def verify_certificate_signature(cls, entity_certificate, ca_certificate):
-        """ Return True if entity_certificate was signed by ca_certificate.
-
-        Args:
-            entity_certificate (str): entity certificate in DER or PEM format.
-            ca_certificate (str): CA certificate in DER or PEM format.
-
-        Return: 
-            bool: True if the ca_certificate validates the entity_certificate.
-        """
-        issuer_public_key = DANE.build_x509_object(ca_certificate).public_key()
-        cert_to_check = DANE.build_x509_object(entity_certificate)
-        if isinstance(issuer_public_key, RSAPublicKey):
-            try:
-                issuer_public_key.verify(cert_to_check.signature,
-                    cert_to_check.tbs_certificate_bytes, padding.PKCS1v15(),
-                    cert_to_check.signature_hash_algorithm)
-            except InvalidSignature:
-                return False
-        elif isinstance(issuer_public_key, EllipticCurvePublicKey):
-            try:
-                issuer_public_key.verify(cert_to_check.signature,
-                    cert_to_check.tbs_certificate_bytes,
-                    cert_to_check.signature_hash_algorithm)
-            except InvalidSignature:
-                return False
-            except UnsupportedAlgorithm:
-                return False
-        else:
-            raise ValueError("Unsupported public key type {}".format(type(issuer_public_key)))
-        return True
-
     @classmethod
     def generate_authority_hostname(cls, identity_name):
         """Return the hostname for an entity's authority server.
@@ -473,7 +388,6 @@ class DANE:
         authority_hostname = ".".join(authority_dns_labels)
         return authority_hostname
 
-
     @classmethod
     def generate_url_for_ca_certificate(cls, authority_hostname, authority_key_id):
         """Return a URL for the identity's ca certificate.
@@ -498,35 +412,71 @@ class DANE:
         return authority_url
 
     @classmethod
-    def get_authority_key_id_from_certificate(cls, certificate):
-        """Return the authorityKeyIdentifier for the certificate.
+    def get_ca_certificates_for_identity(cls, identity_name, certificate, max_levels=100, nsaddr=None, dns_timeout=5):
+        """Return the CA certificates for verifying identity_name.
         
-        
-        Args:
-            certificate (str): Certificate in PEM or DER format.
-        """
-        cert_obj = cls.build_x509_object(certificate)
-        akid = cert_obj.extensions.get_extension_for_class(
-            x509.AuthorityKeyIdentifier
-            ).value.key_identifier
-        return binascii.hexlify(akid, '-').decode()
+        Returns the PEM representation of the CA certificates
+        used for verifying any DANE PKIX-CD certificate 
+        associated with ``identity_name``.
 
-    @classmethod
-    def get_subject_key_id_from_certificate(cls, certificate):
-        """Return the subjectKeyIdentifier for the certificate.
-        
-        
         Args:
+            identity_name (str): DNS name of identity.
             certificate (str): Certificate in PEM or DER format.
+            max_levels (int): Only retrieve this many parent certificates.
+            dns_timeout (int): Timeout in seconds for DNS query.
+
+        Raise:
+            ValueError if no CA certificate is found or the
+                certificate is not parseable.
+
+        Return:
+            list: CA certificates which authenticate identity certificate.
         """
-        cert_obj = cls.build_x509_object(certificate)
-        skid = x509.SubjectKeyIdentifier.from_public_key(cert_obj.public_key())
-        return binascii.hexlify(skid.digest, '-').decode()
+        ca_certs = []
+        authority_hostname = cls.generate_authority_hostname(identity_name)
+        current_cert = PKI.serialize_cert(PKI.build_x509_object(certificate), "PEM")
+        while True:
+            try:
+                parent_ski = PKI.get_authority_key_id_from_certificate(current_cert)
+            except ValueError as err:
+                print("ValueError when parsing AKI from {}: ".format(current_cert, err))
+                raise err
+            ca_certificate_url = cls.generate_url_for_ca_certificate(authority_hostname, parent_ski)
+            presumed_pem = cls.wrap_requests(ca_certificate_url, nsaddr)
+            if not presumed_pem:
+                raise ValueError("CA certificate not found at {}".format(ca_certificate_url))
+            # The following line raises ValueError if it fails to parse.
+            pem = PKI.serialize_cert(PKI.build_x509_object(presumed_pem), "PEM")
+            ca_certs.append(pem)
+            current_cert = pem
+            # This catches self-signed certs.
+            ski = PKI.get_subject_key_id_from_certificate(pem)
+            aki = PKI.get_authority_key_id_from_certificate(pem)
+            if ski == aki:
+                # print("Found root cert, breaking!")
+                break
+            if len(ca_certs) >= max_levels:
+                print("Max iterations reached ({}), breaking!".format(max_levels))
+                break
+        return ca_certs
+    
+    @classmethod
+    def wrap_requests(cls, url, nsaddr=None, dns_timeout=5):
+        """Wrap requests for nameserver override."""
+        parsed = urlparse(url)
+        hostname = parsed.hostname
+        ip_address = cls.get_a_record(hostname, nsaddr, dns_timeout)
+        session = requests.Session()
+        session.mount(url, ForcedIPHTTPSAdapter(dest_ip=ip_address))
+        r = session.get(url, headers={"Host": hostname})
+        return r.content
 
     @classmethod
     def get_ca_certificate_for_identity(cls, identity_name, certificate):
         """Return the CA certificate for verifying identity_name.
         
+        DEPRECATED. USE `get_ca_certificates_for_identity`
+
         Returns the PEM representation of the CA certificate
         used for verifying any DANE PKIX-CD certificate 
         associated with ``identity_name``.
@@ -544,15 +494,14 @@ class DANE:
         """
         authority_hostname = cls.generate_authority_hostname(identity_name)
         try:
-            authority_key_id = cls.get_authority_key_id_from_certificate(certificate)
+            authority_key_id = PKI.get_authority_key_id_from_certificate(certificate)
             ca_certificate_url = cls.generate_url_for_ca_certificate(authority_hostname, authority_key_id)
             r = requests.get(ca_certificate_url)
-            presumed_pem = r.content
             if not r:
                 raise ValueError("CA certificate not found at {}".format(ca_certificate_url))
             # The following line raises ValueError if it fails to parse.
-            x509.load_pem_x509_certificate(presumed_pem, default_backend())
-            return presumed_pem
+            x509.load_pem_x509_certificate(r.content, default_backend())
+            return r.content
         except requests.exceptions.RequestException as err:
             msg = "Error making request: {}".format(err)
             raise ValueError(msg)
@@ -560,5 +509,46 @@ class DANE:
             msg = "Unable to retrieve the authorityKeyID from the certificate."
             raise ValueError(msg)
 
+    @classmethod
+    def process_tlsa(cls, tlsa_record):
+        """Return a dictionary describing the TLSA record's contents.
 
+        Args:
+            tlsa_record (dict): Dictionary describing TLSA record contents.
+
+        Dictionary keys:
+
+            ``tlsa_fields``: TLSA record parsed into a list.
+        
+            ``tlsa_parsed``: A dictionary of parsed TLSA record fields.
+        
+            ``certificate_usage``: Text description of the TLSA field.
+        
+            ``matching_type``: Text description of the TLSA field.
+        
+            ``selector``: Text description of the TLSA field.
+        
+            ``certificate_metadata``: Metadata parsed from the certificate.
+        
+            ``public_key_object``: If the TLSA record contains a public key,
+            this will be the same object as generated by
+            cryptography.hazmat.primitives.serialization.load_der_public_key()
+        
+            ``certificate_object``: If the TLSA record contains a certificate,
+            this will be a cryptography.x509.Certificate object.
+        """
+        tlsa_fields = ["certificate_usage", "selector",
+                       "matching_type", "certificate_association"]
+        cert_association = tlsa_record["certificate_association"]
+        cert_der = PKI.certificate_association_to_der(cert_association)
+        retval = {"certificate_metadata": None, "public_key_object": None}
+        retval["tlsa_fields"] = [tlsa_record[x] for x in tlsa_fields]
+        retval["tlsa_parsed"] = tlsa_record.copy()
+        for target in ["certificate_usage", "matching_type", "selector"]:
+            retval[target] = cls.descr[target][int(tlsa_record[target])]
+        if tlsa_record["matching_type"] == 0:
+            retval["certificate_metadata"] = PKI.get_cert_meta(cert_der)
+            retval["certificate_object"] = PKI.build_x509_object(cert_der)
+            retval["public_key_object"] = retval["certificate_object"].public_key()
+        return retval
 
